package provide Pod 1.1

##  pod.tcl
##
##	pod : an eye-candy toplevel widget
## Copyright (c) 2007 <Irrational Numbers> : <aldo.buratti@tiscali.it> 
##
##
## This library is free software; you can use, modify, and redistribute it
## for any purpose, provided that existing copyright notices are retained
## in all copies and that this notice is included verbatim in any
## distributions.
##
## This software is distributed WITHOUT ANY WARRANTY; without even the
## implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
##


 # WARNING:  for each pod-widget $w, the following widget are created
 #   $w.pod_*   where * is n,s,e,w,nw,ne,se,sw
 # These sub-widgets are RESERVED
 

package require snit
package require Colors

#
# How to use Pod:
#   Read "readme.txt" for detailed info.
#   Sample code in provided in "demo*.tcl".
#


::snit::widgetadaptor Pod {
   typevariable Drag -array {
       DX 0
       DY 0
       enabled false
   }
   #typevariable Drag(DX) is  distance from mouse-x and toplevel-x-corner
   #typevariable Drag(DY) is  distance from mouse-y and toplevel-y-corner

   typevariable Resize -array {
       Xstart 0
       Ystart 0
       W0     0
       H0     0
       cursor {}
   }
   
   typeconstructor {

      # load basic images
     set imageDir [file join [file dirname [info script]] images]
     set imageBaseFile [file join $imageDir podBase.gif]
     set imageBase [image create photo -file $imageBaseFile]

     set dx [expr [image width  $imageBase]/2]
     set dy [expr [image height $imageBase]/2]

     foreach im {nX sX eX wX nw ne se sw} { 
        image create photo Pod::base.${im}
     }
      # images for corners
     Pod::base.nw copy $imageBase  -from 0   0   $dx          $dy
     Pod::base.ne copy $imageBase  -from $dx 0   [expr 2*$dx] $dy
     Pod::base.sw copy $imageBase  -from 0   $dy $dx          [expr 2*$dy]
     Pod::base.se copy $imageBase  -from $dx $dy [expr 2*$dx] [expr 2*$dx]
      # images for sides (resizables)
     Pod::base.nX copy $imageBase  -from $dx 0   [expr $dx+1] $dy
     Pod::base.sX copy $imageBase  -from $dx $dx [expr $dx+1] [expr 2*$dy]
     Pod::base.wX copy $imageBase  -from 0   $dy $dx          [expr $dy+1]
     Pod::base.eX copy $imageBase  -from $dx $dy [expr 2*$dx] [expr $dy+1]
     image delete $imageBase

     image create photo Pod::gripper -file [file join $imageDir grip.gif]     
      # create an 'alternative' version of the 'south-east' image :
      #   'se' plus gripper
     image create photo Pod::base.se_normal
     image create photo Pod::base.se_grip
     Pod::base.se_normal copy Pod::base.se
     Pod::base.se_grip   copy Pod::base.se
     Pod::base.se_grip   copy Pod::gripper -compositingrule overlay
      # Pod::gripper is no more used
     image delete Pod::gripper
  
       # define cursor for resizing
     if { [string equal $::tcl_platform(platform) "windows"] } {
         set Resize(cursor) size_nw_se
     } else {
         set Resize(cursor) bottom_right_corner
     }

       # pseudo class for resizing
      bind Pod.Resizable <ButtonPress-1> \
        [mytypemethod BeginResizeHandler %W %X %Y] 
      bind Pod.Resizable <B1-Motion> \
        [mytypemethod ResizeHandler %W %X %Y] 

      bind Pod <Configure> { 
          %W Resize %w %h user
      }
   }

    # This proc is called whenever user press <Button-1>
    #  on a sub-widget.
    # w is the sub-widget.
    # This proc determines if the sub-widget is "passive" (like a label)
    #  or has its own bindings (even indirects) with <Button-1> (or <Button>.
    #  In this latter case, the dragging action is stopped 
    #  (by setting a flag "Drag(enabled)"
   typemethod BeginMoveHandler {w x y} {   
       #typevariable Drag
      if { [HasItsOwnMotionBinding $w] } {
         set Drag(enabled) false
      } else {
         set topW [winfo toplevel $w]
         set Drag(DX) [expr $x-[winfo x $topW]]
         set Drag(DY) [expr $y-[winfo y $topW]]
	 set Drag(enabled) true
      }
   }

    # true if list L has any item in $args
   proc lsearch_any { L args }  {
      foreach arg $args {
         if { [lsearch -exact $L $arg] != -1 } {
            return true
         }
      }
    return false
   }

   proc HasItsOwnMotionBinding {w} {
      set found false
      set topW [winfo toplevel $w]

      set state normal
      catch { set state [$w cget -state] }

      foreach tag [bindtags $w] {
        if { $tag != $topW } {
           if { 
                $state != "disabled" && 
                [lsearch_any [bind $tag]  <Button> <Button-1> <Motion> <B1-Motion> ] 
              } {
              set found true
              break
           }
        }
      }                                                              
      return $found
   }
     
   typemethod MoveHandler {w x y} {
       #typevariable Drag
      if {$Drag(enabled)} {
         set topW [winfo toplevel $w]
         wm geometry $topW +[expr $x-$Drag(DX)]+[expr $y-$Drag(DY)]
      }
   }


   typemethod BeginResizeHandler {resizerW x y} {
      #typevariable Resize
     set topW [winfo toplevel $resizerW]
     set Resize(H0) [winfo height $topW]
     set Resize(W0) [winfo width $topW]
     set Resize(Xstart) $x
     set Resize(Ystart) $y
   }

      
   typemethod ResizeHandler {resizerW x y} {
      #typevariable Resize
     set topW [winfo toplevel $resizerW]
     set dx [expr $Resize(W0)+$x-$Resize(Xstart)]
     set dy [expr $Resize(H0)+$y-$Resize(Ystart)]
      # do not resize under ...
     set minDX [expr [image width  Pod::base.nw]*2]
     set minDY [expr [image height Pod::base.nw]*2]
     if { $dx < $minDX } { set dx $minDX }
     if { $dy < $minDY } { set dy $minDY }
     wm geometry $topW  ${dx}x${dy}
   }


    # Most of undelying toplevel's options are useless.
    # Some toplevel's options are here intercepted, then passed to toplevel
   option -padx       -default 0 -configuremethod SetPadX
   option -pady       -default 0 -configuremethod SetPadY
   option -background -default #eaeaea  -configuremethod SetBg
    # -bg is an alias for -background
   option -bg         -default #eaeaea  -configuremethod SetBg
   
   option -class -default Pod -readonly true
   
     # note that some options are useless or worst, dangerous
     # e.g -relief -borderwidth -bd
   delegate option * to hull 

    # Pod-specific options 
   option -transparentcolor -default #ff00ff -configuremethod SetTC
   option -resizable        -default true    -configuremethod SetResizable
   
   variable imageDB -array {}
   variable winPod  -array {}
    # Why we use an array for storing the name of the n,s,w, ... labels ?
    #  Why don't you use a simple naming convention 
    #   like the sub-widget 'n' is  ${win}.pod_n  ?
    # Answer:
    #  Because things go wrong when $win is "." ; in such case,
    #   the name of the widget results "..pod_n" and this is invalid
    #   for most commands.
    #  However, note that a command like "label ..pod_n" returns ".pod_n"
    #   and this is correct.
    # For this reason,  using "set winPod(nw) [label ${win}.Pod_nw ...]", 
    #  we can get always a valid wdget name.
    
   variable oldDX 0
   variable oldDY 0
   variable recoloured false


   typemethod adapt {w args} {
      eval $type create $w "*REUSE*" $args
   }
   
   
   constructor {args} {
      if { [lindex $args 0] == "*REUSE*" } {
           installhull $win
            # remove element
           set args [lreplace $args 0 0]
      } else {
           installhull using toplevel
      }


       # substitute Pod with Toplevel
      set idx [lsearch -exact [bindtags $win] Toplevel]
      bindtags $win [lreplace [bindtags $win] $idx $idx Pod]

      catch {
       wm overrideredirect $win 1
       wm attributes $win -transparent $options(-transparentcolor)
      }
            
      foreach im {nX sX eX wX nw ne sw se se_normal se_grip} {
         set imageDB($im) [image create photo]
         $imageDB($im) copy Pod::base.${im}
      }
      foreach place {n s e w} {
         set imageDB($place) [image create photo]
      }

       # create 8 labels for borders and corners
      foreach place {n s e w  nw ne se sw} {
         set winPod($place) [label ${win}.pod_${place} -bd 0 -image $imageDB($place)]
          # borders and corners should react like its own Pod
         bindtags $winPod($place) [list $win]
      }

       # bindings for dragging ...
      bind $win <ButtonPress-1> \
        [mytypemethod BeginMoveHandler %W %X %Y]
      bind $win  <B1-Motion> \
        [mytypemethod MoveHandler %W %X %Y]


       # force grip-image on/off
      $win configure -resizable $options(-resizable)
       # force image placement
      $win SetPadXY 0 0

      $win configurelist $args

      if { ! $recoloured } {
          $win configure -background $options(-background)
      }
   }

   destructor {
       foreach {place img} [array get imageDB] {
           image delete $img
       }
   }

   method SetPadX {option value} {
      $win SetPadXY $value $options(-pady)
      set options($option) $value   
   }
   method SetPadY {option value} {
      $win SetPadXY $options(-padx) $value
      set options($option) $value   
   }
   method SetPadXY {padx pady} {
      set dx [image width   Pod::base.nw]
      set dy [image height  Pod::base.nw]

      $hull configure -padx [expr $padx+$dx] -pady [expr $pady+$dy]
      place $winPod(n)   -anchor s -relx 0.5 -rely 0                    -y [expr -$pady]
      place $winPod(s)   -anchor n -relx 0.5 -rely 1                    -y $pady
      place $winPod(w)   -anchor e -relx 0   -rely 0.5 -x [expr -$padx]
      place $winPod(e)   -anchor w -relx 1   -rely 0.5 -x $padx

      place $winPod(nw) -anchor se -relx 0   -rely 0   -x [expr -$padx] -y [expr -$pady]
      place $winPod(sw) -anchor ne -relx 0   -rely 1   -x [expr -$padx] -y $pady
      place $winPod(ne) -anchor sw -relx 1   -rely 0   -x $padx         -y [expr -$pady]
      place $winPod(se) -anchor nw -relx 1   -rely 1   -x $padx         -y $pady
   }
    
    # set transparent color
   method SetTC {option value} {
      set tc [Colors::TkColor2RRGGBB $value]
       # change color to local images
      foreach place {nw sw ne se se_grip} {
          set newData [string map [list $options($option) $tc] [$imageDB($place) data]]
          $imageDB($place) put $newData
      }
       # note: 'side' images have no transparent pixel  (frame with holes ???)

       # skip error if command is unsupported
      catch {wm attributes $win -transparent $value}

       # store the 'normalized' value
      set options($option) $tc
   }
 
    # set look&feel for resizable Pod
   method SetResizable {option value} {
      if { $value } {
	 $imageDB(se) copy $imageDB(se_grip)
         bindtags $winPod(se) Pod.Resizable
         $winPod(se) configure -cursor $Resize(cursor)
      } else {
	 $imageDB(se) copy $imageDB(se_normal)
         bindtags $winPod(se) [list $win]
         $winPod(se) configure -cursor [$win cget -cursor]
      }
      set options($option) $value
   }

       
    # 3rd parameter: idf equal to "user", then "wm geometry" adjustment is not performed.
    #  Tipically, this proc is bound to a <Configure> event;
    #   in this case the "program" string is appended, meanining that the window won't need
    #   to be explicitely resized.  
    method Resize {DX DY {who program} } {
         # since this is an intensive task, avoid reprocessing 
         #  if DX or DY has not benn changed         
        set d [image height  Pod::base.nw]
        if { $oldDY != $DY &&  $DY > 2*$d } {
             foreach place {e w} {
               image create photo $imageDB($place) ; # reset image
                # +2 pixel added for security ...
               $imageDB($place) copy $imageDB(${place}X) -to 0 0 $d [expr $DY-2*$d+2] 
             }
             set oldDY $DY
        }
        set d [image width Pod::base.nw]
        if { $oldDX != $DX  &&  $DX > 2*$d} {
             foreach place {n s} {
               image create photo $imageDB($place) ; # reset image
               $imageDB($place) copy $imageDB(${place}X) -to 0 0 [expr $DX-2*$d+2] $d 
             }
             set oldDX $DX
        }
        if { $who != "user" } {
           wm geometry $win ${DX}x${DY}
        }
    }

    # this method is called for -bg and -background
   method SetBg {option value} {
      $win Recolour $value
      set options(-background) $value   
      set options(-bg) $value   
   }


    method Recolour {color} {
        # get r,g,b from $color
       foreach {r g b} [Colors::RRGGBB2R_G_B [Colors::TkColor2RRGGBB $color]] {break}
    
        # prepare a color map
       foreach {H S L} [Colors::RgbToHsl $r $g $b] {break}
       set gamma [expr log($L)/log(0.5)]
    
       for {set i 0} {$i < 256} {incr i} {
           set iii [format "#%02x%02x%02x" $i $i $i]
           set L [expr $i/255.0]
           set L [expr pow($L,$gamma)] ; # gamma correction
           foreach { R G B } [Colors::HslToRgb $H $S $L] {break}
           set ColorMap($iii)  [format "#%02x%02x%02x" $R $G $B]
       }
       set ColorMap($options(-transparentcolor)) $options(-transparentcolor)
    
    
       foreach place {nX sX eX wX nw ne sw se_normal se_grip} {
           set y 0
           foreach pixelLine [Pod::base.$place data] {
               set colouredLine {}
               set colouredLine {}
               foreach pixel $pixelLine {
                   lappend colouredLine $ColorMap($pixel)
               }
               $imageDB($place) put [list $colouredLine] -to 0 $y
               incr y
           }
       }
       set rrggbb [format "#%02x%02x%02x" $r $g $b]
       $hull configure -background $rrggbb
        # force grip-image on/off
      $win configure -resizable $options(-resizable)

        # trick: change old coords, 
        #  so that resizable images will be updated with new colors
       set dx $oldDX
       set dy $oldDY
       set oldDX 1
       set oldDY 1
    
       $win Resize $dx $dy user
       set recoloured true
    }

}

